import bcrypt from 'bcrypt';
import jwt from 'jsonwebtoken';
import { from, Observable, of } from 'rxjs';
import { mergeMap } from 'rxjs/operators';
import IUser from '../../interfaces/user';
import AuthResponse from '../../interfaces/auth';
import { getUserByEmail } from '../user/user.functions';
import config from './auth.config';
import { HttpRequest } from '@marblejs/core';

export const authenticateUser = (
  incomingUser: IUser,
) => {
  const user$ = getUserByEmail(incomingUser)
    .pipe(
      mergeMap((user) => {
        if (user.password) {
          return from(bcrypt.compare(incomingUser.password, user.password));
        }

        return of(false);
      }),
    );

  return user$;
};

export const signToken = (user: IUser): AuthResponse => {
  const { _id, email, name } = user;
  const time = new Date().getTime();
  const toExpire = Math.floor((time / 1000) + (Number(config.jwt.expiresIn) * 100));  
  
  const token = jwt.sign(
    {
      _id,
      email,
    },
    config.jwt.secret,
    {
      issuer: config.jwt.issuer,
      algorithm: 'HS256',
      expiresIn: toExpire,
    },
  );

  delete user.password;

  return {
    user: {
      _id: _id as string,
      email,
      name,
    },
    token,
  };
};

export const extractToken = (token: string | undefined): boolean => {
  if (token) {
    const extracted = token.split(' ')[1];
    const verified = jwt.verify(extracted, config.jwt.secret);

    if (verified) {
      return true;
    }

    return false;
  }

  return false;
};
