import {
  HttpError,
  HttpStatus,
  r,
  use,
} from '@marblejs/core';
import {
  throwError,
} from 'rxjs';
import {
  catchError,
  map,
  mergeMap,
  tap,
} from 'rxjs/operators';
import IUser from '../../interfaces/user';
import { authorizeUser$ } from '../auth/auth.middleware';
import {
  getUsers,
  getUser,
  registerUser,
  updateUser,
} from './user.functions';
import userExists$ from './user.middlewares';
import {
  validateRegister,
  validateGetUser,
  validateUpdateUser,
} from './user.validators';

const getUsers$ = r.pipe(
  r.matchPath('/'),
  r.matchType('GET'),
  r.useEffect((req$) => req$.pipe(
    use(authorizeUser$),
    mergeMap(
      getUsers,
    ),
    map((users) => ({
      body: {
        users,
        count: users.length,
      },
    })),
  )),
);

const getUser$ = r.pipe(
  r.matchPath('/:id'),
  r.matchType('GET'),
  r.useEffect((req$) => req$.pipe(
    use(authorizeUser$),
    use(validateGetUser),
    mergeMap(
      (req) => getUser(req.params.id),
    ),
    map((body) => ({
      body,
    })),
  )),
);

const registerUser$ = r.pipe(
  r.matchPath('/register'),
  r.matchType('POST'),
  r.use(userExists$),
  r.useEffect((req$) => req$.pipe(
    use(validateRegister),
    map((req) => <IUser> req.body),
    mergeMap(
      registerUser,
    ),
    catchError((err) => throwError(
      new HttpError(err, HttpStatus.BAD_REQUEST),
    )),
    map((body) => ({
      body,
    })),
  )),
);

const updateUser$ = r.pipe(
  r.matchPath('/'),
  r.matchType('PUT'),
  r.useEffect((req$) => req$.pipe(
    use(authorizeUser$),
    use(validateUpdateUser),
    map((req) => req.body as IUser),
    mergeMap(updateUser),
    catchError((err) => {
      return throwError(
        new HttpError(err, HttpStatus.BAD_REQUEST),
      )
    }),
    map((result) => ({
      body: {
        message: result ? 'User has been saved.' : 'Error saving user.',
      },
    })),
  )),
);

export default [
  getUsers$,
  getUser$,
  registerUser$,
  updateUser$,
];
