import {
  Component,
  EventEmitter,
  Input,
  Output,
  ViewChild,
} from '@angular/core';
import { MatDialog } from '@angular/material/dialog';
import Transformation from 'src/app/models/matrices/transformation';
import Translation3d from 'src/app/models/matrices/translation3d';
import Rotation3d from 'src/app/models/matrices/rotation3d';
import Shear3d from 'src/app/models/matrices/shear3d';
import Scale3d from 'src/app/models/matrices/scale3d';
import Translation2d from 'src/app/models/matrices/translation2d';
import Rotation2d from 'src/app/models/matrices/rotation2d';
import Shear2d from 'src/app/models/matrices/shear2d';
import Scale2d from 'src/app/models/matrices/scale2d';
import {
  TransformationKeys,
  TransformationTypes,
  Axis,
  Axis3d,
  SliderEvent,
  SliderData,
} from 'src/app/models/matrices/transformation.types';
import { TransformationDialogComponent } from './transformation-dialog/transformation-dialog.component';
import { TransformationControlsComponent } from './transformation-controls/transformation-controls.component';
import { Cell } from '../matrix/matrix.types';
import { MatrixComponent } from '../matrix/matrix.component';
import { HelpDialogComponent } from '../help-dialog/help-dialog.component';
import { transformationHelp } from './transformation.help';

@Component({
  selector: 'app-transformation',
  templateUrl: './transformation.component.html',
  styleUrls: ['./transformation.component.scss'],
})
export class TransformationComponent {
  @ViewChild('controls') controls: TransformationControlsComponent;

  @ViewChild('matrix') matrix: MatrixComponent;

  @Output() matrixUpdate = new EventEmitter<Transformation[]>();

  @Input() dimension: 3 | 2 = 3;

  @Input() isCreation = false;

  @Input() section!: 'exploration' | 'application' | 'creation';

  nameToRow: { [name: string]: Cell } = {};

  transformations: Transformation[] = [];

  selectedTransformation: Transformation | null;

  selectedType: TransformationTypes;

  values: SliderData;

  flipChange = false;

  constructor(
    private dialog: MatDialog,
  ) {
    this.transformations.push(
      new Translation3d(),
    );
  }

  openHelpDialog() {
    const data = transformationHelp[this.section];

    if (!data) {
      return;
    }

    this.dialog.open(
      HelpDialogComponent,
      {
        data,
      },
    );
  }

  openDialog(): void {
    const dialogRef = this.dialog.open(
      TransformationDialogComponent,
      {
        data: {
          dimension: this.dimension,
          type: 'Translation',
          axis: 'x',
          max: this.transformations.length,
        },
      },
    );

    dialogRef.afterClosed().subscribe((result) => {
      if (result) {
        const beforeIndex = Math.max(
          this.transformations.indexOf(<Transformation> this.selectedTransformation) - 1,
          0,
        );
        const position = result.before
          ? this.transformations.indexOf(
            <Transformation> this.selectedTransformation,
          ) + 1 : beforeIndex;

        this.addTransformation(
          result.type,
          position,
          result.axis,
        );
      }
    });
  }

  changeTransformation(transformation: Transformation): void {
    this.selectedTransformation = transformation;
    this.values = this.isCreation ? {} : transformation.values;

    const { name } = transformation.constructor;

    switch (name.charAt(0)) {
      case 'R':
        this.selectedType = TransformationTypes.Rotation;
        break;
      case 'S':
        this.selectedType = name.charAt(1) === 'h' ? TransformationTypes.Shear : TransformationTypes.Scale;
        break;
      default:
        this.selectedType = TransformationTypes.Translation;
        break;
    }

    if (this.controls) {
      this.controls.setSliders(this.values);
    }
  }

  removeSelectedTransformation(): void {
    if (!this.selectedTransformation) {
      return;
    }

    const index = this.transformations.indexOf(this.selectedTransformation);
    this.transformations.splice(index, 1);
    this.selectedTransformation = null;
    this.controls?.setSliders({});

    this.matrixUpdate.emit(this.transformations);
  }

  clearTransformations(): void {
    this.selectedTransformation = null;
    this.transformations = [];
    this.controls.setSliders({});
  }

  addTransformation(type: TransformationKeys, position: number, axis?: Axis | Axis3d): void {
    const transformationType = TransformationTypes[type];

    if (this.dimension === 3) {
      switch (transformationType) {
        case TransformationTypes.Rotation:
          console.log('adding rotation');
          this.transformations.splice(
            position,
            0,
            new Rotation3d(axis ?? 'x'),
          );

          return;

        case TransformationTypes.Shear:
          console.log('adding shear');
          this.transformations.splice(
            position,
            0,
            new Shear3d(axis ?? 'x'),
          );

          return;

        case TransformationTypes.Scale:
          console.log('adding scale');
          this.transformations.splice(
            position,
            0,
            new Scale3d(),
          );

          return;

        default:
          console.log('adding translation');
          this.transformations.splice(
            position,
            0,
            new Translation3d(),
          );

          return;
      }
    }

    switch (transformationType) {
      case TransformationTypes.Rotation:
        this.transformations.splice(
          position,
          0,
          new Rotation2d(),
        );
        break;

      case TransformationTypes.Scale:
        this.transformations.splice(
          position,
          0,
          new Scale2d(),
        );
        break;

      case TransformationTypes.Shear:
        this.transformations.splice(
          position,
          0,
          new Shear2d(axis as Axis ?? 'x'),
        );
        break;

      default:
        this.transformations.splice(
          position,
          0,
          new Translation2d(),
        );
        break;
    }
  }

  getTransformationLetter(obj: Object): string {
    return obj.constructor.name.charAt(0);
  }

  isSelected(index: number): boolean {
    const selectedIndex = this.transformations.indexOf(
      <Transformation> this.selectedTransformation,
    );

    return index === selectedIndex;
  }

  getTransformationLetterClass(index: number): string {
    return this.isSelected(index) ? 'transformation-letter selected-transformation-letter' : 'transformation-letter';
  }

  selectTransformation(index: number): void {
    if (!this.isSelected(index)) {
      console.log(`selecting transformation: ${index}`);
      this.changeTransformation(this.transformations[index]);
      return;
    }

    this.selectedTransformation = null;
    this.controls.setSliders({});
  }

  update3dMatrix(name: string, value: number) {
    const transformation = this.selectedTransformation;
    const type = this.selectedType;

    let specific;

    switch (type) {
      case TransformationTypes.Rotation:
        specific = transformation as Rotation3d;
        specific.theta = value;

        break;

      case TransformationTypes.Scale:
        specific = transformation as Scale3d;

        switch (name) {
          case 'y':
            specific.y = value;

            break;

          case 'z':
            specific.z = value;

            break;

          default:
            specific.x = value;
            break;
        }

        break;

      case TransformationTypes.Shear:
        specific = transformation as Shear3d;

        switch (name) {
          case 'x2':
            specific.x2 = value;

            break;

          default:
            specific.x1 = value;
            break;
        }

        break;

      default:
        specific = transformation as Translation3d;

        switch (name) {
          case 'y':
            specific.y = value;

            break;

          case 'z':
            specific.z = value;

            break;

          default:
            specific.x = value;

            break;
        }

        break;
    }
  }

  changeSlider(): void {
    if (!this.isCreation) {
      return;
    }

    const c = this.matrix.selectedCell;
    const values: SliderData = {};
    const name = `(${c.row},${c.col})`;

    values[name] = <number> this.selectedTransformation?.getValue(c.row, c.col);

    this.values = values;

    this.controls?.setSliders(values);
  }

  update2dMatrix(name: string, value: number) {
    const transformation = this.selectedTransformation;
    const type = this.selectedType;
    let specific;

    switch (type) {
      case TransformationTypes.Rotation:
        specific = transformation as Rotation2d;
        specific.theta = value;

        break;

      case TransformationTypes.Scale:
        specific = transformation as Scale2d;

        switch (name) {
          case 'y':
            specific.y = value;

            break;

          default:
            specific.x = value;
            break;
        }

        break;

      case TransformationTypes.Shear:
        specific = transformation as Shear2d;
        specific.x = value;

        break;

      default:
        specific = transformation as Translation2d;

        switch (name) {
          case 'y':
            specific.y = value;

            break;

          default:
            specific.x = value;

            break;
        }

        break;
    }
  }

  sliderReset(data: SliderEvent) {
    this.updateMatrix(data);

    const { name, value } = data;

    this.values[name] = <number> value;

    this.controls.setSliders(this.values);
  }

  updateMatrix(data: SliderEvent) {
    const { name, value } = data;

    if (this.isCreation) {
      const c = this.matrix.selectedCell;

      this.selectedTransformation?.setValue(<number> value, c.row, c.col);
    } else {
      this.selectedTransformation?.set(name, <number> value);
    }

    this.matrix.view = <string[][]> this.selectedTransformation?.viewMatrix;

    this.matrixUpdate.emit(this.transformations);
  }
}
