import {
  AfterViewInit,
  Component,
  ElementRef,
  OnDestroy,
  OnInit,
  ViewChild,
} from '@angular/core';
import {
  fromEvent,
  Subscription,
} from 'rxjs';
import { MatSnackBar } from '@angular/material/snack-bar';
import Viewer from 'src/viewer/viewer';
import Transformation from 'src/app/models/matrices/transformation';

@Component({
  selector: 'app-viewer',
  templateUrl: './viewer.component.html',
  styleUrls: ['./viewer.component.scss'],
})
export class ViewerComponent implements OnInit, AfterViewInit, OnDestroy {
  private _transformations: Transformation[] | undefined;

  @ViewChild('canvasContainer') canvasContainer: ElementRef;

  @ViewChild('canvas') canvasElement: ElementRef<HTMLCanvasElement>;

  get canvas() {
    return this.canvasElement?.nativeElement;
  }

  get gl() {
    return this.viewer.context;
  }

  subs = new Subscription();

  fixSize$ = fromEvent(window, 'resize');

  private viewer: Viewer;

  get scene() {
    return this.viewer?.scene;
  }

  constructor(
    private snackbar: MatSnackBar,
  ) {
  }

  ngOnInit(): void {
    this.subs.add(this.fixSize$.subscribe(() => {
      this.fixCanvasSize();
    }));
  }

  ngAfterViewInit(): void {
    this.viewer = new Viewer(this.canvas);

    this.fixCanvasSize();
    this.setCanvasColor('#e9e9e9');
  }

  ngOnDestroy(): void {
    this.subs.unsubscribe();
  }

  fixCanvasSize(): void {
    const canvasContainer = this.canvasContainer.nativeElement;

    if (!this.canvas) {
      this.snackbar.open('Error opening the viewer.');
      return;
    }

    const width = canvasContainer.getBoundingClientRect().width - 30;

    const height = (3 * width) / 4;

    this.viewer.resizeViewport(width, height);

    this.canvas.setAttribute('width', String(width));
    this.canvas.setAttribute('height', String(height));
  }

  setCanvasColor(color: string) {
    this.canvas.style.setProperty('background-color', color);
  }

  set transformations(t: Transformation[]) {
    this._transformations = t;
  }

  renderScene() {
    this.viewer.render();
  }
}
