import {
  AfterViewInit,
  Component,
  OnDestroy,
  OnInit,
  ViewChild,
} from '@angular/core';
import { MatDialog } from '@angular/material/dialog';
import { ActivatedRoute } from '@angular/router';
import * as moment from 'moment';
import { Moment } from 'moment';
import { fromEvent, interval, Subscription } from 'rxjs';
import { filter, map, take } from 'rxjs/operators';
import { ViewerControlsComponent } from 'src/app/components/viewer/viewer-controls/viewer-controls.component';
import { ViewerComponent } from 'src/app/components/viewer/viewer.component';
import Rotation3d from 'src/app/models/matrices/rotation3d';
import Shear3d from 'src/app/models/matrices/shear3d';
import Scale3d from 'src/app/models/matrices/scale3d';
import Transformation from 'src/app/models/matrices/transformation';
import Transformation3d from 'src/app/models/matrices/transformation3d';
import Translation3d from 'src/app/models/matrices/translation3d';
import User from 'src/app/models/User';
import { UserService } from 'src/app/services/user.service';
import GlBuffer from 'src/viewer/gl/buffer';
import GeometryNode from 'src/viewer/scene/geometry-node';
import Scene from 'src/viewer/scene/scene';
import Shape from 'src/viewer/shapes/shape';
import { FinishedApplicationDialogComponent } from '../finished-application-dialog/finished-application-dialog.component';
import { createScene } from './activity3d.functions';

@Component({
  selector: 'app-activity3d',
  templateUrl: './activity3d.component.html',
  styleUrls: ['./activity3d.component.scss'],
})
export class Activity3dComponent implements OnInit, AfterViewInit, OnDestroy {
  @ViewChild('viewer') viewer: ViewerComponent;

  @ViewChild('controls') controls: ViewerControlsComponent;

  subs = new Subscription();

  lastActive: Moment;

  chosenActivity: 'exploration' | 'application' | 'creation' = 'exploration';

  currentUser: User;

  timerInterval = this.minutesToMs(5);

  timer$ = interval(this.timerInterval);

  constructor(
    public dialog: MatDialog,
    private route: ActivatedRoute,
    private userService: UserService,
  ) { }

  ngOnInit() {
    this.currentUser = this.route.snapshot.data.user;

    if (this.currentUser.problemsSolved === undefined) {
      this.currentUser.problemsSolved = 0;
    }

    if (this.currentUser.explorationTime === undefined) {
      this.currentUser.explorationTime = 0;
    }

    const timerSub = this.timer$.subscribe(() => {
      if (this.chosenActivity !== 'exploration') {
        return;
      }

      const now = moment();
      const then = this.lastActive;
      const diff = now.clone().diff(then, 'minutes');

      if (diff < 5) {
        if (this.currentUser.explorationTime !== undefined) {
          console.log('updating exploration time!');

          this.currentUser.explorationTime += 5;
          return;
        }

        console.log('adding first exploration time');

        this.currentUser.explorationTime = 5;

        console.log(`added exploration time: ${this.currentUser.explorationTime}`);
      }
    });

    const sub1 = fromEvent(window, 'mousemove').subscribe(this.setActive);
    const sub2 = fromEvent(window, 'click').subscribe(this.setActive);
    const sub3 = fromEvent(window, 'keydown').subscribe(this.setActive);
    const sub4 = fromEvent(window, 'mousedown').subscribe(this.setActive);

    this.subs.add(...[timerSub, sub1, sub2, sub3, sub4]);
  }

  updateUser(): void {
    const {
      _id,
      email,
      explorationTime,
      problemsSolved,
    } = this.currentUser;

    const req = {
      id: _id,
      email,
      explorationTime,
      problemsSolved,
    };

    this.userService.update(req)
      .pipe(
        take(1),
      )
      .subscribe((result: User) => console.log(result));
  }

  setActive() {
    if (this.chosenActivity !== 'exploration') {
      return;
    }

    this.lastActive = moment();
  }

  minutesToMs(minutes: number): number {
    return minutes * 60 * 1000;
  }

  ngAfterViewInit() {
    this.setCanvasEvents();

    createScene(this.viewer.gl, <Scene> this.viewer.scene);

    this.viewer.transformations = this.controls.creation.transformations;

    this.renderScene();
  }

  ngOnDestroy() {
    this.subs.unsubscribe();
  }

  setCanvasEvents(): void {
    const { canvas } = this.viewer;

    const coords = {
      x: 0,
      y: 0,
    };

    let mouseDown = false;

    const downSub = fromEvent(canvas, 'mousedown')
      .subscribe(() => {
        mouseDown = true;
      });

    const upSub = fromEvent(canvas, 'mouseup')
      .subscribe(() => {
        mouseDown = false;
      });

    const moveSub = fromEvent(canvas, 'mousemove')
      .pipe(
        map((e: Event) => e as MouseEvent),
        filter(() => mouseDown),
      )
      .subscribe((e: MouseEvent) => {
        const speed = 0.04;
        const x = e.clientX;
        const y = e.clientY;
        const signPhi = Math.sign(x - coords.x);
        const signTheta = Math.sign(y - coords.y);
        const phi = signPhi * speed;
        const theta = signTheta * speed;

        this.viewer?.scene?.activeCamera.rotate([-theta, -phi, 0]);

        coords.x = e.clientX;
        coords.y = e.clientY;
      });

    const zoomSub = fromEvent(canvas, 'wheel')
      .pipe(
        map((e: Event) => e as WheelEvent),
      )
      .subscribe((e: WheelEvent) => {
        e.preventDefault();

        const { deltaY } = e;
        const scaled = -deltaY * 0.01;

        if (!this.viewer.scene?.activeCamera) {
          return;
        }

        const [xp, yp, zp] = this.viewer.scene.activeCamera.position;
        const [xo, yo, zo] = [0.0, 0.0, 0.0];
        const [xd, yd, zd] = [xp - xo, yp - yo, zp - zo];
        const length = Math.sqrt(xd ** 2 + yd ** 2 + zd ** 2);
        const [xn, yn, zn] = [xd / length, yd / length, zd / length];

        const [x, y, z] = [scaled * xn, scaled * yn, scaled * zn];

        this.viewer?.scene?.activeCamera?.translate([-x, -y, -z]);
      });

    this.subs.add(...[moveSub, upSub, downSub, zoomSub]);
  }

  private checkProblemState() {
    const finished = this.viewer.scene?.compareNodeTransformation('main', 'target');

    if (finished) {
      this.removeTargetNode();
      this.resetTransformations();
      this.openFinishedDialog();

      if (this.currentUser?.problemsSolved !== undefined
          && !isNaN(<number> this.currentUser?.problemsSolved)) {
        this.currentUser.problemsSolved++;
      } else {
        this.currentUser.problemsSolved = 1;
      }
    }
  }

  openFinishedDialog() {
    const ref = this.dialog.open(
      FinishedApplicationDialogComponent,
      {
        data: 'Congratulations, you finished!',
      },
    );

    ref.afterClosed().subscribe(() => {
      this.addTargetNode();
      this.updateUser();
    });
  }

  private resetTransformations(): void {
    this.controls[this.chosenActivity].clearTransformations();
    this.updateMatrices([]);
  }

  updateMatrices(matrices: Transformation[]): void {
    const node = <GeometryNode> this.viewer.scene?.getNodeById('main');

    node?.updateTransformationMatrix(<Transformation3d[]> matrices);

    if (this.chosenActivity === 'application') {
      this.checkProblemState();
    }
  }

  private getRandomNumber(start: number, range: number): number {
    return Number((Math.random() * range + start).toFixed(3));
  }

  private getRandomInteger(start: number, range: number): number {
    return Math.floor(Math.random() * range) + start;
  }

  checkRandomTransformation(transformations: Transformation[]): number {
    let rand = this.getRandomInteger(0, 3);

    if (transformations.every((t) => !(t instanceof Translation3d))) {
      return rand;
    }

    while (rand < 1) {
      rand = this.getRandomInteger(0, 3);
    }

    return rand;
  }

  addTargetNode() {
    const mesh = Shape.createCube();
    const buffer = new GlBuffer(this.viewer.gl, mesh, false);

    const nTransformations = Math.floor(Math.random() * 5) + 2;
    const matrices: Transformation[] = new Array(nTransformations).fill(null);

    matrices.forEach((_, i) => {
      const rand = this.getRandomInteger(0, 3);
      let transformation: Transformation3d;

      switch (rand) {
        case 1:
          const theta = this.getRandomNumber(-2 * Math.PI, 2 * Math.PI);
          const axis = this.getRandomNumber(0, 3);

          switch (axis) {
            default:
              transformation = new Rotation3d('x');
              (<Rotation3d> transformation).theta = theta;
              break;
            case 1:
              transformation = new Rotation3d('y');
              (<Rotation3d> transformation).theta = theta;
              break;
            case 2:
              transformation = new Rotation3d('z');
              (<Rotation3d> transformation).theta = theta;
              break;
          }
          break;
        case 2:
          const [xs, ys, zs] = [
            this.getRandomNumber(-6, 6),
            this.getRandomNumber(-6, 6),
            this.getRandomNumber(-6, 6),
          ];

          transformation = new Scale3d();

          (<Scale3d>transformation).x = xs;
          (<Scale3d>transformation).y = ys;
          (<Scale3d>transformation).z = zs;
          break;
        case 3:
          const [x1, x2] = [
            this.getRandomNumber(-6, 6),
            this.getRandomNumber(-6, 6),
          ];

          const ax = this.getRandomNumber(0, 3);

          switch (ax) {
            default:
              transformation = new Shear3d('x');
              (<Shear3d> transformation).x1 = x1;
              (<Shear3d> transformation).x2 = x2;
              break;
            case 1:
              transformation = new Shear3d('y');
              (<Shear3d> transformation).x1 = x1;
              (<Shear3d> transformation).x2 = x2;
              break;
            case 2:
              transformation = new Shear3d('z');
              (<Shear3d> transformation).x1 = x1;
              (<Shear3d> transformation).x2 = x2;
              break;
          }
          break;
        default:
          const [x, y, z] = [
            this.getRandomNumber(-6, 12),
            this.getRandomNumber(-6, 12),
            this.getRandomNumber(-6, 12),
          ];

          transformation = new Translation3d();
          (<Translation3d> transformation).x = x;
          (<Translation3d> transformation).y = y;
          (<Translation3d> transformation).z = z;
          break;
      }

      matrices[i] = transformation;
    });

    this.viewer.scene?.addNode(
      'target',
      mesh,
      [0.0, 10.0, 0.0],
      [0.0, 0.0, 0.0],
      [1.0, 1.0, 1.0],
      [0.0, 0.0, 0.0],
      buffer,
      this.viewer.scene.programs.default,
      'target',
    );

    (<GeometryNode> this.viewer.scene?.getNodeById('target'))?.updateTransformationMatrix(<Transformation3d[]> matrices);
  }

  removeTargetNode() {
    this.viewer.scene?.removeNode('target');
  }

  goToApplication() {
    if (this.chosenActivity === 'exploration') {
      this.updateUser();
    }

    console.log('problems solved: ', this.currentUser.problemsSolved);

    this.removeTargetNode();

    const matrices = this.controls.application.transformations;

    this.updateMatrices(matrices);

    this.addTargetNode();

    this.chosenActivity = 'application';
  }

  goToExploration() {
    if (this.chosenActivity === 'application') {
      this.updateUser();
    }

    this.removeTargetNode();

    const matrices = this.controls.exploration.transformations;

    this.updateMatrices(matrices);

    this.chosenActivity = 'exploration';
  }

  goToCreation() {
    this.updateUser();

    this.removeTargetNode();

    const matrices = this.controls.creation.transformations;

    this.updateMatrices(matrices);

    this.chosenActivity = 'creation';
  }

  renderScene() {
    this.viewer.renderScene();
  }
}
