import Mesh from './mesh';

type BufferData = {
  [dataType: string]: number[];
};

/**
 * class for buffers
 */
class GlBuffer {
  [bufferName: string]: WebGLBuffer;

  gl: WebGL2RenderingContext;

  constructor(
    gl: WebGL2RenderingContext,
    mesh: Mesh,
    pack: boolean,
  ) {
    this.gl = gl;

    let data: BufferData = {};

    if (pack) {
      data = {
        data: GlBuffer.packData(
          mesh.vertices,
          mesh.uvs,
        ),
        indices: mesh.indices,
      };
    } else if (mesh.normals.length) {
      data = {
        vertices: mesh.vertices,
        normals: mesh.normals,
        uvs: mesh.uvs,
        indices: mesh.indices,
      };
    } else {
      data = {
        vertices: mesh.vertices,
        colors: mesh.colors,
        indices: mesh.indices,
      };
    }

    // loop through attrib data
    Object.keys(data).forEach((dataType) => {
      // determine buffer type
      const bufferType = dataType === 'indices'
        ? gl.ELEMENT_ARRAY_BUFFER
        : gl.ARRAY_BUFFER;

      // convert to typed array
      const bufferData = dataType === 'indices'
        ? new Uint16Array(data[dataType])
        : new Float32Array(data[dataType]);

      // bind buffer and transfer data to it
      this.initBuffer(
        gl,
        dataType,
        bufferType,
        bufferData,
      );
    });
  }

  /**
   * Packs the data into one array to fill a buffer
   * @param {number[]} vertices vertex data
   * @param {number[]} colors color data
   *
   * @return {number[]} resulting packed array
   */
  private static packData(
    vertices: number[],
    normals: number[],
    // uvs: number[],
  ): number[] {
    const n = vertices.length / 3;
    const data: number[] = [];

    for (let i = 0; i < n; i++) {
      const index = i * 3;

      // pack vertice data
      data.push(...[
        vertices[index],
        vertices[index + 1],
        vertices[index + 2],
      ]);

      // pack normal data
      data.push(...[
        normals[index],
        normals[index + 1],
        normals[index + 2],
      ]);

      // const colorIndex = i * 4;
      // // pack colors
      // data.push(...[
      //   colors[colorIndex],
      //   colors[colorIndex + 1],
      //   colors[colorIndex + 2],
      //   colors[colorIndex + 3],
      // ]);

      // uvs are 2 so need different index
      // const uvIndex = i * 2;
      // // pack color data
      // data.push(...[
      //   uvs[uvIndex],
      //   uvs[uvIndex + 1],
      // ]);
    }

    // console.log('packed data: ', data);

    return data;
  }

  /**
   * inits the buffers
   * @param {WebGL2RenderingContext} gl webgl context
   * @param {string} dataType name of the buffer
   * @param {GLint} bufferType type of buffer
   * @param {Float32 | Uint16Array} bufferData data to transfer
   *
   * @return {void}
   */
  private initBuffer(
    gl: WebGL2RenderingContext,
    dataType: string,
    bufferType: GLint,
    bufferData: Float32Array | Uint16Array,
  ): void {
    this[dataType] = gl?.createBuffer() as WebGLBuffer;

    gl.bindBuffer(bufferType, this[dataType]);
    gl.bufferData(bufferType, bufferData, gl.STATIC_DRAW);
  }

  cleanBuffer(gl: WebGL2RenderingContext) {
    gl.bindBuffer(gl.ARRAY_BUFFER, this.data);
    gl.bufferData(gl.ARRAY_BUFFER, 1, gl.STATIC_DRAW);
    gl.deleteBuffer(this.data);
    gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER, this.indices);
    gl.bufferData(gl.ELEMENT_ARRAY_BUFFER, 1, gl.STATIC_DRAW);
    gl.deleteBuffer(this.indices);
  }
}

const getStride = (
  sizes: number[],
): number => {
  let stride = 0;

  sizes.forEach((size: number) => {
    stride += size * 4;
  });

  return stride;
};

export const createAttribute = (
  gl: WebGL2RenderingContext,
  buffer: WebGLBuffer,
  attribs: {
    [attribName: string]: {
      size: number;
      location: GLint;
    },
  },
  packed?: boolean,
  debug?: boolean,
): void => {
  gl.bindBuffer(gl.ARRAY_BUFFER, buffer);

  let offset = 0;

  const stride = getStride(Object.values(attribs).map((attrib) => attrib.size));

  if (debug) {
    console.log('stride: ', stride);

    console.log('attribs: ', attribs);
  }

  Object.keys(attribs).forEach((attribName) => {
    const { location, size }: {
      location: GLint;
      size: number;
    } = attribs[attribName];

    gl.enableVertexAttribArray(location);

    gl.vertexAttribPointer(
      location,
      size,
      gl.FLOAT,
      false,
      0,
      offset,
    );

    offset += size * 4;
  });
};

export default GlBuffer;
