type UniformData = {
  [uniformName: string]: {
    uType: string;
    uData: number[];
  };
};

class Program {
  _name: string;

  _gl: WebGL2RenderingContext;

  _program: WebGLProgram | null;

  _shaders: {
    vShader: WebGLShader | null;
    fShader: WebGLShader | null;
  };

  _shaderSrc: {
    [ShaderType: string]: string;
  };

  _attribNames: string[];

  _attribLoc: {
    [attribName: string]: GLint;
  };

  _uniformNames: string[];

  _uniformLoc: {
    [uniformName: string]: WebGLUniformLocation;
  };

  /**
   * constructor
   * @param {WebGL2RenderingContext} gl webgl context
   * @param {string} name  name of the program
   * @param {string[]} attribNames name of attributes
   * @param {string[]} uniformNames name of uniforms
   * @param {string} vSrc source of vertex shader
   * @param {string} fSrc source of fragment sader
   */
  constructor(
    gl: WebGL2RenderingContext,
    name: string,
    attribNames: string[],
    uniformNames: string[],
    vSrc: string,
    fSrc: string,
  ) {
    this._gl = gl;
    this._name = name;
    this._program = this._gl.createProgram();
    this._attribNames = attribNames;
    this._uniformNames = uniformNames;

    this._shaderSrc = {
      vSrc,
      fSrc,
    };

    this._shaders = {
      vShader: gl.createShader(gl.VERTEX_SHADER),
      fShader: gl.createShader(gl.FRAGMENT_SHADER),
    };

    this._attribLoc = {};
    this._uniformLoc = {};
  }

  /**
   * compiles the shaders for this program
   * @param {WebGL2RenderingContext} gl webgl context
   *
   * @returns {void}
   */
  compileShaders(
  ): void {
    if (this._shaders.vShader !== null && this._shaders.fShader !== null) {
      this._gl.shaderSource(this._shaders.vShader, this._shaderSrc.vSrc);
      this._gl.shaderSource(this._shaders.fShader, this._shaderSrc.fSrc);
      this._gl.compileShader(this._shaders.vShader);
      this._gl.compileShader(this._shaders.fShader);

      this.checkShaderCompilation(
        'vertex',
        this._shaders.vShader,
      );

      this.checkShaderCompilation(
        'fragment',
        this._shaders.fShader,
      );
    }
  }

  cleanProgram(gl: WebGL2RenderingContext) {
    gl.deleteShader(this._shaders.vShader);
    gl.deleteShader(this._shaders.fShader);
    gl.deleteProgram(this._program);
  }

  /**
   * Checks the status of a shader compilation
   * @param {string} shaderType vertex/fragment
   * @param {WebGLShader} shader the shader to check
   *
   * @throws {Error} throws an error if there is a problem with shader compilation
   * @returns {void}
   */
  private checkShaderCompilation(
    shaderType: string,
    shader: WebGLShader,
  ): void {
    if (!this._gl.getShaderParameter(shader, this._gl.COMPILE_STATUS)) {
      const errMsg = this._gl.getShaderInfoLog(shader);

      this._gl.deleteShader(shader);

      throw new Error(
        `unable to compile shader of type ${shaderType} on program ${this._name}: ${errMsg}`,
      );
    }
  }

  /**
   * Links the program
   *
   * @throws {Error} throws error if linking fails
   * @returns {void}
   */
  linkProgram(
  ): void {
    if (this._program !== null
          && this._shaders.vShader !== null
          && this._shaders.fShader !== null) {
      this._gl.attachShader(this._program, this._shaders.vShader);
      this._gl.attachShader(this._program, this._shaders.fShader);
      this._gl.linkProgram(this._program);

      // check if saul goodman with program linking
      if (!this._gl.getProgramParameter(this._program, this._gl.LINK_STATUS)) {
        const errMsg = this._gl.getProgramInfoLog(this._program);

        this._gl.deleteProgram(this._program);

        throw new Error(
          `unable to link program ${this._name}: ${errMsg}`,
        );
      }

      this.getLocations();
    }
  }

  /**
   * Gets the locations of the attributes / uniforms
   * @param {WebGL2RenderingContext} gl webgl context
   *
   * @returns {void}
   */
  private getLocations(
  ): void {
    this._attribNames.forEach((attribName) => {
      if (this._program !== null) {
        this._attribLoc[attribName] = this._gl.getAttribLocation(this._program, attribName);
      }
    });

    this._uniformNames.forEach((uniformName) => {
      if (this._program !== null) {
        this._uniformLoc[uniformName] = <WebGLUniformLocation>
          this._gl.getUniformLocation(this._program, uniformName);
      }
    });
  }

  activateProgram() {
    this._gl.useProgram(this._program);
  }

  sendUniforms(uniformData: UniformData, transpose = false, debug?: boolean) {
    if (debug) {
      console.log('sending uniforms...');
    }

    Object.keys(uniformData).forEach((uniformName) => {
      const location = this._uniformLoc[uniformName];
      const type = uniformData[uniformName].uType;
      const data = uniformData[uniformName].uData;

      if (debug) {
        console.log(`${uniformName}, ${type}, ${location}`, data);
      }

      switch (type) {
        case 'int':
          this._gl.uniform1i(location, data[0]);
          break;

        case 'float':
          this._gl.uniform1f(location, data[0]);
          break;

        case 'vec3':
          this._gl.uniform3fv(location, new Float32Array(data));
          break;

        case 'vec4':
          this._gl.uniform4fv(location, new Float32Array(data));
          break;

        case 'mat3':
          this._gl.uniformMatrix3fv(location, transpose, new Float32Array(data));
          break;

        case 'mat4':
          this._gl.uniformMatrix4fv(location, transpose, new Float32Array(data));
          break;

        default:
          break;
      }
    });
  }

  get name() {
    return this._name;
  }
}

export default Program;
