import { fromEvent } from 'rxjs';
import { take } from 'rxjs/operators';

export default class Texture {
  private _gl: WebGL2RenderingContext;

  private _texture: WebGLTexture;

  private _url: string;

  private _config: {
    level: number;
    internalFormat: number;
    width: number;
    height: number;
    border: number;
    srcFormat: number;
    srcType: number;
    pixel: Uint8Array;
  };

  constructor(
    gl: WebGL2RenderingContext,
    url: string,
  ) {
    this._gl = gl;
    this._texture = gl.createTexture() as WebGLTexture;
    this._url = url;

    this._config = {
      level: 0,
      internalFormat: this._gl.RGBA,
      width: 1,
      height: 1,
      border: 0,
      srcFormat: this._gl.RGBA,
      srcType: this._gl.UNSIGNED_BYTE,
      pixel: new Uint8Array([0, 0, 255, 255]),
    };
  }

  loadTexture(): void {
    this.loadImage();

    this._gl.texImage2D(
      this._gl.TEXTURE_2D,
      this._config.level,
      this._config.internalFormat,
      this._config.width,
      this._config.height,
      this._config.border,
      this._config.srcFormat,
      this._config.srcType,
      this._config.pixel,
    );
  }

  private loadImage() {
    const gl = this._gl;
    const img = new Image();
    const img$ = fromEvent(img, 'load').pipe(take(1));

    img$.subscribe({
      next: () => {
        gl.bindTexture(gl.TEXTURE_2D, this._texture);

        gl.texImage2D(
          gl.TEXTURE_2D,
          this._config.level,
          this._config.internalFormat,
          this._config.srcFormat,
          this._config.srcType,
          img,
        );

        if (this.powerOf2(img.width) && this.powerOf2(img.height)) {
          gl.generateMipmap(gl.TEXTURE_2D);
        } else {
          gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_S, gl.CLAMP_TO_EDGE);
          gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_T, gl.CLAMP_TO_EDGE);
          gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, gl.LINEAR);
        }
      },
      error: (err) => console.error(err),
    });

    img.onload = () => {
    };

    img.src = this._url;

    return img;
  }

  private powerOf2(val: number): boolean {
    return (val & (val - 1)) === 0;
  }

  get texture(): WebGLTexture {
    return this._texture;
  }
}
