import { mat4 } from 'gl-matrix';
import Transformation3d from 'src/app/models/matrices/transformation3d';
import Translation3d from 'src/app/models/matrices/translation3d';
import Rotation3d from 'src/app/models/matrices/rotation3d';
import Scale3d from 'src/app/models/matrices/scale3d';
import Mesh from '../gl/mesh';
import GlBuffer, { createAttribute } from '../gl/buffer';
import SceneNode from './scene-node';
import Program from '../gl/program';
import Texture from '../gl/texture';

class GeometryNode extends SceneNode {
  private _mesh: Mesh;

  private _vaoCount: number;

  private _program: Program;

  private _buffer: GlBuffer;

  private _texture: Texture;

  private _drawMode: number;

  constructor(
    id: string,
    position: number[],
    rotation: number[],
    scale: number[],
    refPoint: number[],
    mesh: Mesh,
    vaoCount: number,
    buffer: GlBuffer,
    program: Program,
    texture?: Texture,
  ) {
    super(id, position, rotation, scale, refPoint);
    this._mesh = mesh;
    this._vaoCount = vaoCount;
    this._buffer = buffer;
    this._program = program;

    if (texture) {
      this._texture = texture;
    }
  }

  get mesh() {
    return this._mesh;
  }

  get vaoCount() {
    return this._vaoCount;
  }

  get program() {
    return this._program;
  }

  set texture(t: Texture) {
    this._texture = t;
  }

  get texture() {
    return this._texture;
  }

  clearTransformation(): void {
    mat4.identity(this._transformation);
  }

  updateTransformationMatrix(transformations: Transformation3d[] | undefined): void {
    mat4.identity(this._transformation);

    transformations?.forEach((t: Transformation3d) => {
      if (t instanceof Translation3d) {
        this.position = [t.x, t.y, t.z];
      } else if (t instanceof Rotation3d) {
        switch (t.axis) {
          case 'y':
            this.rotation[1] = t.theta;
            break;
          case 'z':
            this.rotation[2] = t.theta;
            break;
          default:
            break;
        }
      } else if (t instanceof Scale3d) {
        this.scale = [t.x, t.y, t.z];
      }

      const m = t.matrix;
      const cM = m.columnMajor;

      const matrix = mat4.fromValues(
        cM[0],
        cM[1],
        cM[2],
        cM[3],
        cM[4],
        cM[5],
        cM[6],
        cM[7],
        cM[8],
        cM[9],
        cM[10],
        cM[11],
        cM[12],
        cM[13],
        cM[14],
        cM[15],
      );

      mat4.multiply(this._transformation, this._transformation, matrix);
    });
  }

  // updateTransformation(parentTransformation: mat4 = mat4.create()) {
  // }

  render(gl: WebGL2RenderingContext, debug?: boolean) {
    createAttribute(
      gl,
      this._buffer.vertices,
      {
        position: {
          size: 3,
          location: this._program._attribLoc.position,
        },
      },
      debug,
    );

    if (this.mesh.normals.length) {
      createAttribute(
        gl,
        this._buffer.normals,
        {
          normal: {
            size: 3,
            location: this._program._attribLoc.normal,
          },
        },
      );
    }

    if (this.mesh.uvs.length) {
      createAttribute(
        gl,
        this._buffer.uvs,
        {
          uv: {
            size: 2,
            location: this._program._attribLoc.uv,
          },
        },
        debug,
      );
    }

    if (this.mesh.colors.length) {
      createAttribute(
        gl,
        this._buffer.colors,
        {
          color: {
            size: 4,
            location: this._program._attribLoc.color,
          },
        },
        debug,
      );
    }

    this._program.sendUniforms({
      mMat: {
        uType: 'mat4',
        uData: this._transformation as number[],
      },
    });

    if (debug) {
      console.log('position location: ', this._program._attribLoc.position);
      console.log('color location: ', this._program._attribLoc.color);

      console.log(`vaoCount: ${this._vaoCount}`);

      console.log('mesh: ', this._mesh);

      console.log('node: ', this);

      console.log('data: ', this._buffer.data);
    }

    gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER, this._buffer.indices);

    const mode = this._mesh.drawMode ? this._mesh.drawMode : gl.TRIANGLES;

    gl.drawElements(
      mode,
      this._vaoCount,
      gl.UNSIGNED_SHORT,
      0,
    );
  }

  get buffer(): GlBuffer {
    return this._buffer;
  }
}

export default GeometryNode;
