import { mat4, vec3 } from 'gl-matrix';

export default class SceneNode {
  protected _position: number[];

  protected _rotation: number[];

  protected _scale: number[];

  protected _refPoint: number[];

  _transformation: mat4;

  protected _children: SceneNode[];

  protected _id: string;

  constructor(
    id: string,
    position: number[],
    rotation: number[],
    scale: number[],
    refPoint: number[],
    children?: SceneNode[],
  ) {
    this._id = id;
    this._position = position;
    this._rotation = rotation;
    this._scale = scale;
    this._refPoint = refPoint;
    this._children = children === undefined ? [] : children;
    this._transformation = mat4.create();
    mat4.identity(this._transformation);
  }

  public updateTransformation(
    parentTransformation: mat4 = mat4.create(),
  ): void {
    mat4.identity(parentTransformation);

    mat4.translate(
      this._transformation,
      parentTransformation,
      vec3.fromValues(
        this._position[0],
        this._position[1],
        this._position[2],
      ),
    );
    mat4.rotateX(
      this._transformation,
      this._transformation,
      this.rotation[0],
    );
    mat4.rotateY(
      this._transformation,
      this._transformation,
      this.rotation[1],
    );
    mat4.rotateZ(
      this._transformation,
      this._transformation,
      this.rotation[2],
    );
    mat4.scale(
      this._transformation,
      this._transformation,
      vec3.fromValues(
        this._scale[0],
        this._scale[1],
        this._scale[2],
      ),
    );

    this._children?.forEach((child) => {
      child.updateTransformation(this._transformation);
    });
  }

  get id(): string {
    return this._id;
  }

  get position(): number[] {
    return this._position;
  }

  set position(p: number[]) {
    this._position = p;
  }

  get rotation(): number[] {
    return this._rotation;
  }

  set rotation(r: number[]) {
    this._rotation = r;
  }

  get scale(): number[] {
    return this._scale;
  }

  set scale(s: number[]) {
    this._scale = s;
  }

  get refPoint(): number[] {
    return this._refPoint;
  }

  set refPoint(r: number[]) {
    this._refPoint = r;
  }

  get children() {
    return this._children;
  }

  get transformation() {
    return this._transformation;
  }

  addChild(child: SceneNode) {
    this._children.push(child);
  }
}
