import { mat4, vec4 } from 'gl-matrix';
import Program from '../gl/program';
import SceneNode from './scene-node';
import GeometryNode from './geometry-node';
import CameraNode from './camera-node';
import Mesh from '../gl/mesh';
import GlBuffer from '../gl/buffer';
import Texture from '../gl/texture';

import { ProjectionMode } from '../constants';
import PROJECTION_CONFIG from '../config';
import Programs from '../types/program';
import ProgramNodeList from '../types/node';

class Scene {
  private _gl: WebGL2RenderingContext;

  private _programs: Programs = {};

  private _rootNode: SceneNode;

  private _nodes: SceneNode[];

  private _cameras: CameraNode[] = [];

  private _activeCamera: CameraNode | null = null;

  private _viewMatrix: mat4 | null = null;

  private _projectionMatrix: mat4;

  private _projectionMode: ProjectionMode = ProjectionMode.PERSPECTIVE;

  private _textures: {
    [id: string]: Texture,
  } = {};

  constructor(
    gl: WebGL2RenderingContext,
    nodes: SceneNode[],
    programs: Programs,
    aspectRatio?: number,
  ) {
    this._gl = gl;
    this._nodes = [...nodes];
    this._programs = programs;
    this._projectionMatrix = mat4.create();
    mat4.identity(this._projectionMatrix);
    const viewport = gl.getParameter(gl.VIEWPORT);
    const aspect = aspectRatio || viewport[2] / viewport[3];

    mat4.perspective(
      this._projectionMatrix,
      PROJECTION_CONFIG.fov,
      aspect,
      PROJECTION_CONFIG.zNear,
      PROJECTION_CONFIG.zFar,
    );

    this._rootNode = this.createRootNode();

    this._textures.cube = new Texture(this._gl, '/assets/textures/purple-cube-texture.png');
    this._gl.bindTexture(this._gl.TEXTURE_2D, this._textures.cube.texture);
    this._textures.cube.loadTexture();

    this._textures.target = new Texture(this._gl, '/assets/textures/target-cube-texture.png');
    this._gl.bindTexture(this._gl.TEXTURE_2D, this._textures.target.texture);
    this._textures.target.loadTexture();
  }

  private createRootNode(): SceneNode {
    const node = new SceneNode(
      'rootNode',
      [0, 0, 0],
      [0, 0, 0],
      [0, 0, 0],
      [0, 0, 0],
      this._nodes,
    );

    return node;
  }

  addTexture(id: string, url: string) {
    this._textures[id] = new Texture(this._gl, url);
  }

  addNode(
    name: string,
    mesh: Mesh,
    position: number[],
    rotation: number[],
    scale: number[],
    refPoint: number[],
    buffer: GlBuffer,
    program: Program,
    texture?: string,
  ): string {
    const id = name;

    const node = new GeometryNode(
      id,
      position,
      rotation,
      scale,
      refPoint,
      mesh,
      mesh.vaoCount,
      buffer,
      program,
      this._textures[texture as string] ? this._textures[texture as string] : undefined,
    );

    this._nodes.push(node);

    return id;
  }

  removeNode(id: string): void {
    const node = this._nodes.find((n) => n.id === id);

    if (!node) {
      return;
    }

    const index = this._nodes.indexOf(node);

    this._nodes.splice(index, 1);
  }

  updateScene() {
    this._nodes.forEach((node) => {
      node.updateTransformation();
    });
  }

  getNodeById(id: string): SceneNode | undefined {
    return this._nodes.find((node) => node.id === id);
  }

  compareNodeTransformation(id1: string, id2: string): boolean {
    const [n1, n2] = [this.getNodeById(id1), this.getNodeById(id2)];

    if (!n1 || !n2) {
      return false;
    }

    return this.compareNodeTransformations(<GeometryNode> n1, <GeometryNode> n2);
  }

  compareNodeTransformations(
    compareNode: GeometryNode,
    targetNode: GeometryNode,
  ): boolean {
    const targetVertices = targetNode.mesh.vertices;
    const targetTransformation = targetNode.transformation;
    const movingVertices = compareNode.mesh.vertices;
    const compareTransformation = compareNode.transformation;

    for (let i = 0; i < targetVertices.length; i += 3) {
      const targetVector = vec4.fromValues(
        targetVertices[i],
        targetVertices[i + 1],
        targetVertices[i + 2],
        1,
      );

      const compareVector = vec4.fromValues(
        movingVertices[i],
        movingVertices[i + 1],
        movingVertices[i + 2],
        1,
      );

      vec4.transformMat4(targetVector, targetVector, targetTransformation);
      vec4.transformMat4(compareVector, compareVector, compareTransformation);

      const distance = vec4.dist(targetVector, compareVector);

      if (distance > 0.02) {
        return false;
      }
    }

    return true;
  }

  setPerspective(aspect: number, fov: number, zNear?: number, zFar?: number): void {
    mat4.identity(this._projectionMatrix);

    mat4.perspective(
      this._projectionMatrix,
      fov,
      aspect,
      zNear || 0.1,
      zFar || 100,
    );
  }

  getNodesByProgram(debug?: boolean): ProgramNodeList {
    return this._nodes.reduce((prev, curr) => {
      if (curr instanceof GeometryNode) {
        const programName = curr.program.name;

        if (debug) {
          console.log(programName);
        }

        if (programName in prev) {
          prev[programName].push(curr);
        } else {
          prev[programName] = [curr];
        }
      }

      return prev;
    }, {} as ProgramNodeList);
  }

  renderScene(debug?: boolean) {
    const nodesByProgram = this.getNodesByProgram();
    this._activeCamera?.calculateViewMatrix();

    let vMat = this._activeCamera?.viewMatrix;

    if (!vMat) {
      vMat = mat4.create();
      mat4.identity(vMat);
    }

    Object.entries(nodesByProgram).forEach(([programName, nodes]) => {
      const program = this._programs[programName];

      if (debug) {
        console.log(programName);
      }

      program.activateProgram();

      program.sendUniforms(
        {
          pMat: {
            uType: 'mat4',
            uData: <number[]> this._projectionMatrix,
          },
          vMat: {
            uType: 'mat4',
            uData: <number[]> vMat,
          },
        },
        false,
        debug,
      );

      nodes.forEach((node) => {
        if (debug) {
          console.log('node to render: ', node);
        }

        if (node.texture) {
          this._gl.activeTexture(this._gl.TEXTURE0);
          this._gl.bindTexture(this._gl.TEXTURE_2D, node.texture.texture);

          program.sendUniforms({
            uSampler: {
              uType: 'int',
              uData: [0],
            },
          });
        }

        program.sendUniforms(
          {
            mMat: {
              uType: 'mat4',
              uData: <number[]> node.transformation,
            },
          },
          false,
          debug,
        );

        if (node.mesh.normals.length) {
          const nMat = mat4.create();

          mat4.multiply(nMat, <mat4> vMat, node.transformation);
          mat4.invert(nMat, nMat);
          mat4.transpose(nMat, nMat);

          program.sendUniforms(
            {
              nMat: {
                uType: 'mat4',
                uData: <number[]> nMat,
              },
            },
          );
        }

        node.render(this._gl, debug);
      });
    });
  }

  addCamera(camera: CameraNode) {
    this._cameras.push(camera);
  }

  set activeCamera(camera: CameraNode) {
    this._activeCamera = camera;
  }

  get activeCamera() {
    return <CameraNode> this._activeCamera;
  }

  get nodes() {
    return this._nodes;
  }

  get programs() {
    return this._programs;
  }
}

export default Scene;
