import Program from './gl/program';
import Scene from './scene/scene';
import Renderer from './render';
import {
  Shaders,
  AttribNames,
  UniformNames,
} from './constants';
import SceneNode from './scene/scene-node';
import GeometryNode from './scene/geometry-node';
import CameraNode from './scene/camera-node';
import PROJECTION_CONFIG from './config';

export default class Viewer {
  private _canvas: HTMLCanvasElement;

  private _gl: WebGL2RenderingContext;

  private _programs: {
    [name: string]: Program;
  } = {};

  private _scene: Scene | null = null;

  private _renderer: Renderer | null = null;

  /**
   * Constructor
   * @param {HTMLCanvasElement} canvas canvas to draw on
   */
  constructor(
    canvas: HTMLCanvasElement,
  ) {
    this._canvas = canvas;
    this._gl = <WebGL2RenderingContext> this._canvas.getContext('webgl2');
    this.init();
  }

  resizeViewport(width: number, height: number) {
    this._gl.viewport(0, 0, width, height);

    const aspect = width / height;

    const { fov } = PROJECTION_CONFIG;

    this._scene?.setPerspective(aspect, fov);
  }

  /**
   * Gets the canvas of the viewer
   *
   * @returns {HTMLCanvasElement} the canvas element
   */
  get canvas(): HTMLCanvasElement {
    return this._canvas;
  }

  get context(): WebGL2RenderingContext {
    return this._gl;
  }

  /**
   * Creates a program and the scene
   *
   * @returns {void}
   */
  private init(): void {
    this._programs.default = new Program(
      this._gl,
      'default',
      AttribNames.DEFAULT,
      UniformNames.DEFAULT,
      Shaders.DEFAULT.VERT,
      Shaders.DEFAULT.FRAG,
    );

    this._programs.default.compileShaders();
    this._programs.default.linkProgram();

    this._programs.line = new Program(
      this._gl,
      'line',
      AttribNames.LINE,
      UniformNames.LINE,
      Shaders.LINE.VERT,
      Shaders.LINE.FRAG,
    );

    this._programs.line.compileShaders();
    this._programs.line.linkProgram();

    const {
      width,
      height,
    } = this._canvas.getBoundingClientRect();

    const aspect = width / height;

    this._scene = new Scene(
      this._gl,
      [],
      this._programs,
      aspect,
    );

    const camera = new CameraNode(
      'mainCamera',
      [0.0, 0.0, 10.0],
      [0.0, 0.0, 0.0],
      [1.0, 1.0, 1.0],
      [0.0, 0.0, 0.0],
    );

    this._scene.addCamera(camera);
    this._scene.activeCamera = camera;

    this._renderer = new Renderer(
      this._gl,
      Object.values(this._programs),
      this._scene,
      width,
      height,
    );
  }

  cleanResources() {
    this._scene?.nodes.forEach((node: SceneNode) => {
      if (node instanceof GeometryNode) {
        node?.buffer?.cleanBuffer(this._gl);
      }
    });

    Object.values(this._programs).forEach((p) => p?.cleanProgram(this._gl));
  }

  get scene() {
    return this._scene;
  }

  render() {
    this._renderer?.runRenderLoop(() => {
    });
  }
}
